/* GNU Chess 5.0 - book.c - book code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
     lukas@debian.org
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "common.h"
#include "book.h"

#define MAXMOVES 200
#define MAXMATCH 100

static int bookcnt;
static HashType posshash[MAXMOVES];
static boolean_t sdbg = TRUE;

/*
 * New stuff
 */
extern FILE *gdbg_fp;
int BookPVSQuery(HashType hkey, char *move, int score, int *new_score);
int TreePVSQuery(HashType hkey, char *move, int score, int *new_score);
static void hi_merge_clusters(void);
static void tr_print_stats(void);
boolean_t tr_query(HashType hkey, char *move);

/*
 * This is the only authoritative variable telling us
 * whether a book has been allocated or not. (Other
 * parts may mess around with bookloaded.)
 */
static int book_allocated = 0;

/*
 * The last byte of magic_str should be the version
 * number of the format, in case we have to change it.
 *
 * Format 0x01 had an index field which is now gone.
 *
 * Format 0x03 uses an additional size header, which 
 * comes directly after the magic string, and has the
 * number of entries as a big-endian encoded uint32_t
 * number. 
 */

#define MAGIC_LENGTH 5

static const char magic_str[] = "\x42\x23\x08\x15\x03";


static int check_magic(FILE *f)
{
  char buf[MAGIC_LENGTH];
  int r;

  r = fread(&buf, 1, MAGIC_LENGTH, f);
  return (r == MAGIC_LENGTH &&
	  memcmp(buf, magic_str, MAGIC_LENGTH) == 0);
}

static int write_magic(FILE *f)
{
  if (MAGIC_LENGTH != fwrite(&magic_str, 1, MAGIC_LENGTH, f)) {
    return BOOK_EIO;
  } else {
    return BOOK_SUCCESS;
  }
}

/* Write and read size information for the binary book */

static int write_size(FILE *f, uint32_t size)
{
  unsigned char sizebuf[4];
  int k;

  for (k = 0; k < 4; k++) {
    sizebuf[k] = (size >> ((3-k)*8)) & 0xff;
  }
  if (1 == fwrite(&sizebuf, sizeof(sizebuf), 1, f)) {
    return BOOK_SUCCESS;
  } else {
    return BOOK_EIO;
  }
}

/* Returns 0 if some read error occurs */

static uint32_t read_size(FILE *f)
{
  unsigned char sizebuf[4];
  uint32_t size = 0;
  int k;

  if (1 != fread(&sizebuf, sizeof(sizebuf), 1, f)) {
    return 0;
  }
  for (k = 0; k < 4; k++) {
    size = (size << 8) | sizebuf[k];
  }
  return size;
}

/*
 * We now allocate memory for the book dynamically,
 * according to the size field in the header of the binary
 * book. However, during book building the following value
 * is used.
 */

#define MAX_DIGEST_BITS 20

static int digest_bits;

#define DIGEST_SIZE (1UL << digest_bits)
#define DIGEST_MASK (DIGEST_SIZE - 1)

static struct hashtype {
  uint16_t wins;
  uint16_t losses;
  uint16_t draws;
  HashType key;
  int      tree;
} *bookpos;

static inline int is_empty(uint32_t index)
{
  return 
    bookpos[index].key    == 0 &&
    bookpos[index].wins   == 0 &&
    bookpos[index].draws  == 0 &&
    bookpos[index].losses == 0;
}

/*
 * Initial hash function, relies on the quality of the lower
 * bits of the 64 bit hash function
 */
#define DIGEST_START(i,key) \
  ((i) = (key) & DIGEST_MASK)

/*
 * See if there is already a matching entry
 */
#define DIGEST_MATCH(i,the_key) \
  ((the_key) == bookpos[i].key)

/*
 * See if the entry is empty
 */
#define DIGEST_EMPTY(i) is_empty(i)

/*
 * Check for collision
 */
#define DIGEST_COLLISION(i,key) \
  (!DIGEST_MATCH(i,key) && !DIGEST_EMPTY(i))

/*
 * The next macro is used in the case of hash collisions.
 * We use double hashing with the higher bits of the key.
 * In order to have the shift relatively prime to the hash
 * size, we OR by 1.
 */
#define DIGEST_NEXT(i,key) \
  ((i) = ((i) + (((key) >> digest_bits) | 1)) & DIGEST_MASK)


/* Mainly for debugging purposes */
static int bookhashcollisions = 0;

/*
 * This is considered to be the limit for the hash, I chose
 * 95% because it is Monday... No, I am open for suggestions
 * for the right value, I just don't know.
 */
#define DIGEST_LIMIT (0.95 * DIGEST_SIZE)

/*
 * This is the record as it will be written in the binary
 * file in network byte order. HashType is uint64_t. To
 * avoid endianness and padding issues, we do not read or
 * write structs but put the values in an unsigned char array.
 */

static unsigned char buf[2+2+2+8];

/* Offsets */
static const int wins_off   = 0;
static const int losses_off = 2;
static const int draws_off  = 4;
static const int key_off    = 6;

static void buf_to_book(void)
{
  HashType key;
  uint32_t i;

  key = ((uint64_t)buf[key_off] << 56)
    | ((uint64_t)buf[key_off+1] << 48)
    | ((uint64_t)buf[key_off+2] << 40)
    | ((uint64_t)buf[key_off+3] << 32)
    | ((uint64_t)buf[key_off+4] << 24)
    | ((uint64_t)buf[key_off+5] << 16)
    | ((uint64_t)buf[key_off+6] << 8)
    | ((uint64_t)buf[key_off+7]);
  /*
   * This is an infinite loop if the hash is 100% full,
   * but other parts should check that this does not happen.
   */
  for (DIGEST_START(i, key);
       DIGEST_COLLISION(i, key);
       DIGEST_NEXT(i, key))
    /* Skip */
    bookhashcollisions++;

  bookpos[i].wins   += (buf[wins_off]   << 8) | buf[wins_off  +1];
  bookpos[i].draws  += (buf[draws_off]  << 8) | buf[draws_off +1];
  bookpos[i].losses += (buf[losses_off] << 8) | buf[losses_off+1];
  bookpos[i].key = key;
  return;
}

static void book_to_buf(uint32_t index)
{
  int k;

  for (k=0; k<2; k++) {
    buf[wins_off + k]   = ((bookpos[index].wins)   >> ((1-k)*8)) & 0xff;
    buf[draws_off + k]  = ((bookpos[index].draws)  >> ((1-k)*8)) & 0xff;
    buf[losses_off + k] = ((bookpos[index].losses) >> ((1-k)*8)) & 0xff;
  }
  for (k=0; k<8; k++) {
    buf[key_off + k] = ((bookpos[index].key) >> ((7-k)*8)) & 0xff;
  }
}

static int compare(const void *aa, const void *bb)
{
  const leaf *a = aa;
  const leaf *b = bb;
  
  if (b->score > a->score) return(1);
  else if (b->score < a->score) return(-1);
  else return(0);
}

/*
 * Reads an existing binary book from f. The header must
 * already be skipped, when you call this function. The
 * variable digest_bits must be set to the correct value
 * before calling this function. If any book was allocated
 * before, it will be lost.
 */
static int read_book(FILE *f)
{
  if (book_allocated) {
    free(bookpos);
    book_allocated = 0;
  }
  bookpos = calloc(DIGEST_SIZE, sizeof(struct hashtype));
  if (bookpos == NULL) {
    return BOOK_ENOMEM;
  }
  book_allocated = 1;
  bookcnt = 0;
  bookhashcollisions = 0;
  while ( 1 == fread(&buf, sizeof(buf), 1, f) ) {
    buf_to_book();
    bookcnt++;
  }
  return BOOK_SUCCESS;
}
  
/* 
 * Return values are defined in common.h
 */

int BookBuilderOpen(void)
{
  FILE *rfp, *wfp;
  int res;

  if ((rfp = fopen(BOOKRUN,"rb")) != NULL) {
    printf("Opened existing book!\n");
    if (!check_magic(rfp)) {
      fprintf(stderr, 
	      "File %s does not conform to the current format.\n"
	      "Consider rebuilding your book.\n",
	      BOOKRUN);
      return BOOK_EFORMAT;
    }
    /*
     * We have to read the size header, but in book building we
     * use the maximum-sized hash table, so we discard the value.
     */
    digest_bits = MAX_DIGEST_BITS;
    read_size(rfp);
    res = read_book(rfp);
    fclose(rfp);
    if (res != BOOK_SUCCESS) {
      fclose(rfp);
      return res;
    }
    printf("Read %d book positions\n", bookcnt);
    printf("Got %d hash collisions\n", bookhashcollisions);
  } else {
    wfp = fopen(BOOKRUN,"w+b");
    if (wfp == NULL) {
      fprintf(stderr, "Could not create %s file: %s\n",
		 BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    if (write_magic(wfp) != BOOK_SUCCESS) {
      fprintf(stderr, "Could not write to %s: %s\n",
	      BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    if (fclose(wfp) != 0) {
      fprintf(stderr, "Could not write to %s: %s\n",
	      BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    printf("Created new book %s!\n", BOOKRUN);
    rfp = fopen(BOOKRUN, "rb");
    if (rfp == NULL) {
      fprintf(stderr, "Could not open %s for reading: %s\n",
	      BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    digest_bits = MAX_DIGEST_BITS;
    /* We use read_book() here only to allocate memory */
    if (read_book(wfp) == BOOK_ENOMEM) {
      return BOOK_ENOMEM;
    }
  }
  return BOOK_SUCCESS;
}

/*
 * Store the position and results of last search
 * if and only if in the first 10 moves of the game.
 * This routine is called after the computer makes its
 * most recent move. Lastly, after the 10th move, on 
 * the 11th move, store the data out to the running file.
 */
/*
 * NOTE: Before you build a book, you have to call
 * BookBuilderOpen() now, after building it BookBuilderClose()
 * in order to actually write the book to disk.
 */

int BookBuilder(short result, uint8_t side)
{
  uint32_t i;
  
  /* Only first BOOKDEPTH moves */
  if (GameCnt > BOOKDEPTH) {
    return BOOK_EMIDGAME;
  }

  CalcHashKey();
  for (DIGEST_START(i, HashKey);
       ; 
       DIGEST_NEXT(i, HashKey)) {
    if (DIGEST_MATCH(i, HashKey)) {
      existpos++;
      break;
    } else if (DIGEST_EMPTY(i)) {
      if (bookcnt > DIGEST_LIMIT)
	return BOOK_EFULL;
      bookpos[i].key = HashKey;
      newpos++;
      bookcnt++;
      break;
    } else {
      bookhashcollisions++;
    }
  }
  if (side == white) {
    if (result == R_WHITE_WINS)
      bookpos[i].wins++;
    else if (result == R_BLACK_WINS)
      bookpos[i].losses++;
    else if (result == R_DRAW)
      bookpos[i].draws++;
  } else {
    if (result == R_WHITE_WINS)
      bookpos[i].losses++;
    else if (result == R_BLACK_WINS)
      bookpos[i].wins++;
    else if (result == R_DRAW)
      bookpos[i].draws++;
  }
  return BOOK_SUCCESS;
}

char *book_file = NULL;
int ntrees_needed = 0;
extern void C_BookPGNReadFromFile (const char *file);

int BookBuilderClose(void)
{
  /*
   * IMHO the following part needs no temporary file.
   * If two gnuchess invocations try to write the same
   * file at the same time, this goes wrong anyway.
   * Please correct me if I am wrong. If you generate
   * a temporary file, try to generate it in some writable
   * directory.
   */
  FILE *wfp;
  unsigned int i;
  int errcode = BOOK_SUCCESS;

  wfp = fopen(BOOKRUN, "wb");
  if (wfp == NULL) {
    errcode = BOOK_EIO;
    goto bailout_noclose;
  }
  if (write_magic(wfp) != BOOK_SUCCESS) {
    errcode = BOOK_EIO;
    goto bailout;
  }
  if (write_size(wfp, bookcnt) != BOOK_SUCCESS) {
    errcode = BOOK_EIO;
    goto bailout;
  }
  for (i = 0; i < DIGEST_SIZE; i++) {
    if (!is_empty(i)) {
      book_to_buf(i);
      if (1 != fwrite(&buf, sizeof(buf), 1, wfp)) {
	errcode = BOOK_EIO;
	goto bailout;
      }
    }
  }
  printf("Got %d hash collisions\n", bookhashcollisions);


 bailout:
  if (fclose(wfp) != 0) {
    errcode = BOOK_EIO;
  }


  if (flags & NTREE) {
    ntrees_needed = 1;
    C_BookPGNReadFromFile(book_file);
	tr_print_stats();

	/*
	 * Start building hierarchical ntrees now!
	 */
	if (flags & HIER) {
		hi_merge_clusters();
	}
  }



 bailout_noclose:
  free(bookpos);
  book_allocated = 0;

  /* Let BookQuery allocate again */
  bookloaded = 0;

  return errcode;
}

/*
 * Return values are defined in common.h
 * NOTE: This function now returns BOOK_SUCCESS on
 * success, which is 0. So all the possible callers
 * should adjust to this. (At present it is only called
 * in iterate.c.) It used to return 1 on success before.
 */

int BookQuery(int BKquery)
{
  /*
   * BKquery denotes format; 0 text, 1 engine 
   * In general out put is engine compliant, lines start with a blank
   * and end with emtpy line
   */
  int i,j,k,icnt = 0, mcnt, found, tot, maxdistribution;
  int matches[MAXMATCH] ;
  leaf m[MAXMOVES];
  leaf pref[MAXMOVES];
  struct {
    uint16_t wins;
    uint16_t losses;
    uint16_t draws;
  } r[MAXMOVES];
  FILE *rfp = NULL;
  leaf *p;
  short side,xside,temp;
  uint32_t booksize;
  int res;

  if (bookloaded && book_allocated) {
  	if ((flags & UPDATESCORE) ||
		(flags & TUPDATESCORE)) { 
		printf("SKIP book lookups initially\n");
		return BOOK_ENOMOVES;
	}
  }




  if (bookloaded && !book_allocated) {
    /* Something failed during loading the book */
    return BOOK_ENOBOOK;
  }
  if (!bookloaded) {
    char const * const *booktry;

    bookloaded = 1;
    for (booktry = bookbin; *booktry ; booktry++) {
      if (!(flags & XBOARD)) {
	fprintf(ofp, "Looking for opening book in %s...\n", *booktry);
      }
      rfp = fopen(*booktry, "rb");
      /* XXX: Any further error analysis here?  -- Lukas */
      if (rfp == NULL) continue;
      if (!(flags & XBOARD))
	fprintf(ofp,"Read opening book (%s)...\n", *booktry);
      if (!check_magic(rfp)) {
	fprintf(stderr, 
		" File %s does not conform to the current format.\n"
		" Consider rebuilding it.\n\n",
		*booktry);
	fclose(rfp);
	rfp = NULL;
      } else break; /* Success, stop search */
    }
    if (rfp == NULL) {
      /* If appropriate print error */
      if (!(flags & XBOARD) || BKquery == 1)
        fprintf(ofp," No book found.\n\n");
      return BOOK_ENOBOOK;
    }
    if (!(flags & XBOARD)) {
      fprintf(ofp, "Loading book from %s.\n", *booktry);
    }
    /*
     * The factor 1.06 is the minimum additional amount of
     * free space in the hash
     */
    booksize = read_size(rfp) * 1.06;
    for (digest_bits = 1; booksize; booksize >>= 1) {
      digest_bits++;
    }
    res = read_book(rfp);
    if (res != BOOK_SUCCESS) {
      return res;
    }
    if (!(flags & XBOARD))
      fprintf(ofp,"%d hash collisions... ", bookhashcollisions);
  }      

  mcnt = -1;
  side = board.side;
  xside = 1^side;
  TreePtr[2] = TreePtr[1];
  GenMoves(1);
  FilterIllegalMoves (1);
  for (p = TreePtr[1]; p < TreePtr[2]; p++) {
    MakeMove(side, &p->move);
    m[icnt].move = p->move;
    posshash[icnt] = HashKey;
    icnt++;
    UnmakeMove(xside,&p->move);
  }
  for (i = 0; i < icnt; i++) {
    for (DIGEST_START(j,posshash[i]);
	 !DIGEST_EMPTY(j);
	 DIGEST_NEXT(j, posshash[i])) {
      if (DIGEST_MATCH(j, posshash[i])) {
	found = 0;
	for (k = 0; k < mcnt; k++) 
	  if (matches[k] == i) {
	    found = 1;
	    break;
	  }
	/* Position must have at least some wins to be played by book */
	if (!found) {
	  matches[++mcnt] = i;
	  pref[mcnt].move = m[matches[mcnt]].move;
	  r[i].losses = bookpos[j].losses;
	  r[i].wins = bookpos[j].wins;
	  r[i].draws = bookpos[j].draws;
	
	  /* by percent score starting from this book position */
	  
	  pref[mcnt].score = m[i].score = 
	    100*(r[i].wins+(r[i].draws/2))/
	    (MAX(r[i].wins+r[i].losses+r[i].draws,1)) + r[i].wins/2;
	  
	}
	if (mcnt >= MAXMATCH) {
	  fprintf(ofp," Too many matches in book.\n\n");
	  goto fini;
	}
	break; /* If we found a match, the following positions can not match */
      }
    }
  }

fini:  

  if (bookloaded && book_allocated) {
  	if ((flags & UPDATESCORE) ||
		(flags & TUPDATESCORE)) { 
		printf("SKIP book lookups initially\n");
		return BOOK_ENOMOVES;
	}
  }


  if (!(flags & XBOARD) || BKquery == 1)
  {
    fprintf(ofp," Opening database: %d book positions.\n",bookcnt);
    if (mcnt+1 == 0) {
      fprintf(ofp," In this position, there is no book move.\n\n");
    } else if (mcnt+1 == 1) {
      fprintf(ofp," In this position, there is one book move:\n");
    } else {
      fprintf(ofp," In this position, there are %d book moves:\n", mcnt+1);
    }
  }
  /* No book moves */
  if (mcnt == -1) {
    return BOOK_ENOMOVES;
  }


  k = 0;
  if (mcnt+1) {
    if ( !(flags & XBOARD) || BKquery == 1 ) {
      for (i = 0; i <= mcnt; i++) {
	if (!(flags & XBOARD) || BKquery == 1 ) {
	  SANMove(m[matches[i]].move,1);
          tot = r[matches[i]].wins+r[matches[i]].draws+r[matches[i]].losses;
	  fprintf(ofp," %s(%2.0f/%d/%d/%d) ",SANmv,
		100.0*(r[matches[i]].wins+(r[matches[i]].draws/2.0))/tot,
		r[matches[i]].wins,
		r[matches[i]].losses,
		r[matches[i]].draws);
          if ((i+1) % 4 == 0) fputc('\n',ofp);
	}
      }
      if (!(flags & XBOARD) || BKquery == 1 )
        if (i % 4 != 0) fprintf(ofp," \n \n");
    }
    if (bookmode == BOOKRAND) {
      k = rand();
      k = k % (mcnt+1);
      RootPV = m[matches[k]].move;
      if (!(flags & XBOARD)) {
        printf("\n(Random picked move #%d %s%s from above list)\n",k,
	  algbr[FROMSQ(RootPV)],algbr[TOSQ(RootPV)]);
        tot = r[matches[k]].wins+r[matches[k]].draws+r[matches[k]].losses;
        if (tot != 0)
          printf("B p=%2.0f\n",
	   100.0*(r[matches[k]].wins+r[matches[k]].draws)/tot);
        else
          printf("p=NO EXPERIENCES\n");
      }
    } else if (bookmode == BOOKBEST) {
      qsort(&pref,mcnt+1,sizeof(leaf),compare);
      RootPV = pref[0].move;
    } else if (bookmode == BOOKWORST) {
      qsort(&pref,mcnt+1,sizeof(leaf),compare);
      RootPV = pref[mcnt].move;
    } else if (bookmode == BOOKPREFER) {
      qsort(&pref,mcnt+1,sizeof(leaf),compare);
      for (i = 0; i <= mcnt; i++) {
	if (!(flags & XBOARD) || BKquery == 1) {
	  SANMove(pref[i].move,1);
          printf(" %s(%d) ",SANmv,pref[i].score);
	}
	m[i].move = pref[i].move;
	if (!(flags & XBOARD) || BKquery == 1) 
          if ((i+1) % 8 == 0) fputc('\n',ofp);
      }
      if (!(flags & XBOARD) || BKquery == 1)
	if (i % 8 != 0) fprintf(ofp," \n \n");
        temp = (bookfirstlast > mcnt+1 ? mcnt+1 : bookfirstlast);
      /* Choose from the top preferred moves based on distribution */
      maxdistribution = 0;
      for (i = 0; i < temp; i++)
        maxdistribution += pref[i].score;
      /* Do not play moves that have only losses! */
      if (maxdistribution == 0) 
	return BOOK_ENOMOVES;
      k = rand() % maxdistribution;
      maxdistribution = 0;
      for (i = 0; i < temp; i++) {
	maxdistribution += pref[i].score;
	if (k >= maxdistribution - pref[i].score &&
	    k < maxdistribution)
	{
	  k = i;
	  RootPV = m[k].move;
	  break;
	}
      }
    }
  }
  return BOOK_SUCCESS;
}
/*
 * Code for N-ARY trees
 */

#define MAX_MOVES_PER_TYPE 16

struct cboard {
	HashType     b_hkey;
	char         b_cmove[SANSZ];  /* Character move format */
	int          b_move;     /* This move lead to this new board */
	short        b_side;     /* Side played this move now */
	uint16_t     b_wins;     /* Number of wins on making this move */
	uint16_t     b_losses;   /* Number of losses on making this move */
	float        b_ratio;    /* Ratio using wins and losses */
	short        b_index;    /* This is where the next child will go */
	short        b_tree;     /* My tree number */
	int          b_depth;    /* depth of this node */
	int          b_nelems;   /* Number of elements in this tree */
	int          b_game;     /* My game number from book */
	struct cboard *b_parent;
	struct cboard *b_sibling;
	struct cboard *n_bp[MAX_MOVES_PER_TYPE];  /* Next set of boards */
};

typedef struct cboard cboard_t;
static int tree_last_updated;

#define MAX_TREES    1024
#define HI_MAX_DEPTH    3  /* Used only for hierarchical clusters */
#define TR_DN_MAX       10  /* Used only if depthnodes is set */

static cboard_t *ntrees[MAX_TREES];
static int cindex;
static int tree_num = 1;
static int bgame;
static cboard_t *tr_last_move_black_entry;

void tr_new_book_move(char *move, int nmove);
static boolean_t tr_check_hashkey(char *move, uint32_t *bpos_index);
static void tr_build_new_tree(int tr_num, char *move,
		int nmove, uint32_t bpos_index);
static cboard_t *tr_find_entry(cboard_t *root, HashType hkey);
static void tr_add_entry(cboard_t *pbp, char *move, int nmove, uint32_t bpos_index);
static boolean_t tr_add_to_existing_tree(char *move, int nmove,
	uint32_t bpos_index, int *depth);
static void tr_print_tree_stats(int c);
static int tr_dist(int i, int j);
static int tr_wins_depth(cboard_t *root, int depth);

/*
 * We just read a new book move.  
 * Check if this HashKey is in any tree
 * Check if HashKey is in the bookpos and check if 
 * the tree is updated yet.  If not, then update
 * the hashkey and the start building the tree
 */
void
tr_new_book_move(char *move, int nmove)
{
	cboard_t *cbp = NULL;
	int first;
	int index = 0;
	cboard_t *bp;
	uint32_t bpos_index = 0;

	//printf("tr_new_book_move:\n");
	//printf("New book move %s. GameCnt %d\n", move, GameCnt);

	if (GameCnt > BOOKDEPTH) {
		//printf("MAX REACHED\n");
		return;
	}
	if (tree_num >= MAX_TREES) {
		//printf("MAX REACHED\n");
		return;
	}

	if (GameCnt/2 + 1 == 1 && (1^board.side) == white) {
		bgame++;
		printf("\n\nREADING NEW BOARD GAME %d\n\n", bgame);
	}

	if (tr_check_hashkey(move, &bpos_index) == FALSE) {
		/*
		 * Not in any tree.
		 */
		if (GameCnt/2 + 1 == 1 && (1^board.side) == white) {
			/*
			 * Reading a brand new game which is not in 
			 * any tree.  Start a new tree.
			 */
			tr_build_new_tree(tree_num, move, nmove, bpos_index);
			tree_last_updated = tree_num;
			tree_num++;
		} else {
			/*
			 * Check if we can add to any existing tree
			 */
			int depth = 0;
			boolean_t added = FALSE;

			added = tr_add_to_existing_tree(move, nmove,
				bpos_index, &depth);
			if (added == FALSE) {
				if (flags & HIER) {
					if (depth == HI_MAX_DEPTH) {
						tr_build_new_tree(tree_num, move, nmove, bpos_index);
						tree_last_updated = tree_num;
						tree_num++;
					}
				}
				if (flags & DEPTHNODES) {
					if (depth == TR_DN_MAX-1) {
						tr_build_new_tree(tree_num, move, nmove, bpos_index);
						tree_last_updated = tree_num;
						tree_num++;
					}
				}
			}
		}
	} else {
		/*
		 * The entry is already in some tree.  Nothing to do since
		 * the wins and losses have been taken already taken from
		 * bookpos.
		 */
		return;
	}
	//printf("tr_new_book_move: DONE\n");
	return;
}

/* 
 * Return TRUE if already in some tree
 */
static boolean_t
tr_check_hashkey(char *move, uint32_t *bpos_index)
{
	uint32_t i;

	//printf("tr_check_hashkey: Move %s HashKey 0x%x\n", move, HashKey);
	for (DIGEST_START(i, HashKey); ;DIGEST_NEXT(i, HashKey)) {
		if (DIGEST_MATCH(i, HashKey)) {
			if (bookpos[i].tree == 0) {
				//printf("Match found move %s Hashkey 0x%x. No tree\n", move, HashKey);
				*bpos_index = i;
				return FALSE;
			} else {
				//printf("Entry %s already in tree %d\n", move, bookpos[i].tree);
				*bpos_index = i;
				return TRUE;
			}
		}
	}
	printf("SHOULD NEVER BE HERE\n");
	return FALSE;
}

static void
tr_build_new_tree(int tr_num, char *move, int nmove, uint32_t bpos_index)
{
	cboard_t *cbp = NULL;
	leaf *p;
	char      wb[32];
	int index = cindex++;
	//printf("tr_build_new_tree:\n");
	if (index > MAX_TREES) {
		printf("MAX ntrees reached. Cannot add new ntrees.  Skipped move %s\n", move);
		return;
	}

	cbp = ntrees[index] = (cboard_t *) malloc(sizeof(cboard_t));
	bzero(cbp, sizeof(cboard_t));

	cbp->b_hkey = bookpos[bpos_index].key;
	strcpy(cbp->b_cmove, move);
	cbp->b_move    = nmove;
	cbp->b_side    = 1^board.side;
	cbp->b_wins    = bookpos[bpos_index].wins;
	cbp->b_losses  = bookpos[bpos_index].losses;
	cbp->b_ratio   = (float)(cbp->b_wins - cbp->b_losses) / (float)(cbp->b_wins+cbp->b_losses);
	cbp->b_tree = tr_num;
	cbp->b_parent  = NULL;
	cbp->b_sibling = NULL;
	cbp->b_depth   = 0;
	cbp->b_game    = bgame;
	ntrees[tr_num-1]->b_nelems++;

	bookpos[bpos_index].tree = tr_num;

	bzero(wb, 32);
	if (cbp->b_side == white) {
		strcpy(wb, "white");
	} else {
		strcpy(wb, "black");
	}
	printf("NEW N-ARY TREE : tree %d, move %s, side %s, wins %d, losses %d, ratio %.2f\n",
		cbp->b_tree,
		cbp->b_cmove,
		wb,
		cbp->b_wins,
		cbp->b_losses,
		cbp->b_ratio);
	printf("tr_build_new_tree:done\n");
	return;
}

/*
 * This move was found in the book pos index (bpos_index).
 * It is not in any tree.
 * Add it as the child of PrevHashKey
 *
 * The PrevHashKey should be found in the tree_last_updated
 */
static boolean_t
tr_add_to_existing_tree(char *move, int nmove, uint32_t bpos_index,
	int *depth)
{
	int c;
	cboard_t *pbp; 
	*depth = 0;

	//printf("tr_add_to_existing_tree:\n");
	for (c = 0; c < MAX_TREES; c++) {
		if (ntrees[c] == NULL) {
			printf("Cannot add move %s to any tree. tree_num = %d\n", move, c);
			return FALSE;
		}

		/*
		 * Instead of looking in all the ntrees, just look
		 * at the tree from bookpos for the PrevHashKey
		 */

		pbp = tr_find_entry(ntrees[c], PrevHashKey);
		if (pbp) {
			if (flags & HIER) {
				if (pbp->b_depth == HI_MAX_DEPTH) {
					/*
				 	 * Max elements already.
				 	 */
					 *depth = pbp->b_depth;
					return FALSE;
				}
			}
			if (flags & DEPTHNODES) {
				if (pbp->b_depth == TR_DN_MAX-1) {
					/*
				 	 * Max elements already.
				 	 */
					 *depth = pbp->b_depth;
					return FALSE;
				}
			}
			tr_add_entry(pbp, move, nmove, bpos_index);
			return TRUE;
		}
	}
	printf("Could not find 0x%x in any tree\n", PrevHashKey);
	return FALSE;
}

static cboard_t *
tr_find_entry(cboard_t *root, HashType phkey)
{
	cboard_t *current = root;
	if (current->b_hkey == phkey) {
		return current;
	}
    while (current != NULL) {
		if (current->n_bp[0]) {
			current = current->n_bp[0]; 
			if (current && current->b_hkey == phkey) {
				return current;
			}
		} else {
			/*
			 * Leaf node
			 */
			while (current->b_sibling == NULL && current != root) {
				current = current->b_parent;
				if (current && current->b_hkey == phkey) {
					return current;
				}
			}
			current = current->b_sibling;
			if (current && current->b_hkey == phkey) {
				return current;
			}
		}
	}
	return NULL;
}

/*
 * We pass the parent information and the new move information.
 * Add the new entry
 */
static void
tr_add_entry(cboard_t *pbp, char *move, int nmove, uint32_t bpos_index)
{
	cboard_t  *bp;
	cboard_t  *tbp;
	cboard_t  *fbp;
	char      wb[32];
	int index = pbp->b_index++;
	//printf("tr_add_entry:\n");
	if (index >= MAX_MOVES_PER_TYPE) {
		printf("tr_add_entry: No more children allowed.  Skipped move %s\n",
			move);
		return;
	}
	bp = pbp->n_bp[index] = (cboard_t *) malloc(sizeof(cboard_t));
	bzero(bp, sizeof(cboard_t));

	bp->b_hkey = bookpos[bpos_index].key;
	strcpy(bp->b_cmove, move);
	bp->b_move    = nmove;
	bp->b_side    = 1^board.side;
	bp->b_wins    = bookpos[bpos_index].wins;
	bp->b_losses  = bookpos[bpos_index].losses;
	bp->b_ratio   = (float)(bp->b_wins - bp->b_losses) / (float)(bp->b_wins+bp->b_losses);
	bp->b_tree = pbp->b_tree;
	bp->b_game = bgame;
	bp->b_parent  = pbp;
	bp->b_sibling = NULL;
	bp->b_depth   = pbp->b_depth + 1;
	ntrees[bp->b_tree-1]->b_nelems++;
	if (index > 0) {
		cboard_t *prev_sbp;  /* Prev sibling */
		prev_sbp = pbp->n_bp[index - 1];
		prev_sbp->b_sibling = bp;
	}
	bookpos[bpos_index].tree = pbp->b_tree;

	bzero(wb, 32);
	if (bp->b_side == white) {
		strcpy(wb, "white");
	} else {
		strcpy(wb, "black");
	}
	printf("NEW CHILD ADDED: tree %d, parent %s, child %s, side %s, child num %d, wins %d, losses %d, ratio %.2f\n",
		bp->b_tree,
		pbp->b_cmove,
		bp->b_cmove,
		wb,
		index,
		bp->b_wins,
		bp->b_losses,
		bp->b_ratio);
	printf("tr_add_entry:DONE\n");

	/*
	 * Swap the children if necessary now.
	 */
	if (index > 0 && pbp->n_bp[index]->b_wins > pbp->n_bp[0]->b_wins) {
		tbp              = pbp->n_bp[index];
		pbp->n_bp[index] = pbp->n_bp[0];
		pbp->n_bp[0]     = tbp;
		pbp->n_bp[0]->b_sibling = pbp->n_bp[1];
		pbp->n_bp[index-1]->b_sibling = pbp->n_bp[index];
		pbp->n_bp[index]->b_sibling = NULL;
	}
	return;
}

/*
 * The user has made a move.  The computer needs to make a next move.
 *
 * Find if there is an entry in our ntrees for this hash key.
 * If there is one, the leftmost child will be the best move.
 * Set the rootpv to the left most child and set the timeout.
 *
 * If there is a last move black entry, check only its children
 * to match any of the white moves made by user.  If it does
 * not match, then we don't have that entry in our tree.
 * If it matches, use the white entry's leftmost black child
 */
boolean_t
tr_query(HashType hkey, char *move)
{
	int c;
	cboard_t *pbp; 
	int i;

	printf("\ntr_query: Looking for move = %s\n", move);

	if (GameCnt >= BOOKDEPTH) {
		return FALSE;
	}

	if (tr_last_move_black_entry) {
		cboard_t *pbp = tr_last_move_black_entry;
		for (i = 0; i < MAX_MOVES_PER_TYPE; i++) {
			cboard_t *wbp = pbp->n_bp[i];
			if (wbp == NULL) {
				/*
				 * No more children
				 */
				printf("No more children for %s\n", move);
				tr_last_move_black_entry = NULL;
				return FALSE;
			}
			if (wbp->b_hkey == hkey) {
				/*
				 * This wbp is the current white move.  We need
				 * its left most child as the next black move.
				 */
				cboard_t *bp = wbp->n_bp[0];
				printf("N-ARY TREE LOOKUP SUCCESS: tree %d, parent %s, child %s, wins %d\n",
					bp->b_tree,
					pbp->b_cmove,
					bp->b_cmove,
					bp->b_wins);
	  			RootPV = bp->b_move;
				tr_last_move_black_entry = bp;
				return TRUE;
			}
		}
	}

	for (c = 0; c < MAX_TREES; c++) {
		if (ntrees[c] == NULL) {
			printf("Cannot find move %s in any tree\n", move);
			return FALSE;
		}
		pbp = tr_find_entry(ntrees[c], hkey);
		if (pbp) {
			/*
			 * Found the entry
			 */
			cboard_t *bp = pbp->n_bp[0];
			if (bp) {
				printf("N-ARY TREE LOOKUP SUCCESS: tree %d, parent %s, child %s, wins %d\n",
				bp->b_tree,
				pbp->b_cmove,
				bp->b_cmove,
				bp->b_wins);
	  			RootPV = bp->b_move;
				tr_last_move_black_entry = bp;
				return TRUE;
			} else {
				printf("This hashkey is a leaf node! %s\n", move);
				return FALSE;
			}
		}
	}
	printf("Could not find move %s 0x%x in any tree\n", move, hkey);
	return FALSE;
}

static int tr_nodes_depth(cboard_t *root, int depth);
static int tr_avg_nodes_depth(int depth);

static void
tr_print_stats(void)
{
	cboard_t *pbp; 
	int i;
	int j;

	for (i = 0; i < MAX_TREES; i++) {
		if (ntrees[i] == NULL) {
			break;
		}
	}
	printf("\nTOTAL NUMBER OF NARY TREES CREATED: %d\n", i);
	for (i = 0; i < MAX_TREES && ntrees[i]; i++) {
		tr_print_tree_stats(i);
	}

	if (flags & DEPTHNODES) {
		for (i = 0; i < MAX_TREES && ntrees[i]; i++) {
			//printf("\nTREE = %d\n", i);
			for (j = 0; j < TR_DN_MAX; j++) {
				tr_nodes_depth(ntrees[i], j);
			}
		}
		for (j = 0; j < TR_DN_MAX; j++) {
			tr_avg_nodes_depth(j);
		}
	}
	return;
}

static void
tr_print_tree_stats(int c)
{	
	int depth = 0;
	int count = 0;
	cboard_t *root = ntrees[c];
	cboard_t *current = ntrees[c];

    while (current != NULL) {
		count++;
		depth = MAX(depth, current->b_depth);
		if (current->n_bp[0]) {
			current = current->n_bp[0]; 
			if (current) {
				depth = MAX(depth, current->b_depth);
			}
		} else {
			/*
			 * Leaf node
			 */
			while (current->b_sibling == NULL && current != root) {
				current = current->b_parent;
				if (current) {
					depth = MAX(depth, current->b_depth);
				}
			}
			current = current->b_sibling;
			if (current) {
				depth = MAX(depth, current->b_depth);
			}
		}
	}
	if (sdbg) {
		printf("TREE = %d, root = %s, depth = %d, num elems = %d\n", ntrees[c]->b_tree, root->b_cmove, depth, root->b_nelems);
	}
	return;
}

/*
 * Return the distance between ntrees[i] and ntrees[j]
 */
static int
tr_dist(int i, int j)
{
	int      dist = 0;
	cboard_t *bp1 = ntrees[i];
	cboard_t *bp2 = ntrees[j];
	int depth[HI_MAX_DEPTH];
	int k;

	if (bp1 == NULL || bp2 == NULL) {
		printf("tr_dist.  Null bp1 or null bp2\n");
		return 0;
	}
	for (k = 0; k < HI_MAX_DEPTH; k++) {
		depth[k] = 0;
	}
	for (k = HI_MAX_DEPTH - 1; k >= 0 ; k--) {
		if (k == HI_MAX_DEPTH - 1) {
			depth[k] = abs(tr_wins_depth(bp1, k) - tr_wins_depth(bp2, k));
		} else {
			depth[k] = abs(tr_wins_depth(bp1, k) - tr_wins_depth(bp2, k)) 
						+ depth[k + 1];
		}
		dist += depth[k];
	}
	//printf("Distance between %s:%s %d\n", bp1->b_cmove, bp2->b_cmove,  dist);
	return dist;
}

static int
tr_wins_depth(cboard_t *root, int depth)
{
	cboard_t *current = root;
	int wins = 0;
	if (depth == 0) {
		return current->b_wins;
	}
    while (current != NULL) {
		if (current->n_bp[0]) {
			current = current->n_bp[0]; 
			if (current && current->b_depth == depth) {
				wins += current->b_wins;
			}
		} else {
			/*
			 * Leaf node
			 */
			while (current->b_sibling == NULL && current != root) {
				current = current->b_parent;
			}
			current = current->b_sibling;
			if (current && current->b_depth == depth) {
				wins += current->b_wins;
			}
		}
	}
	if (wins == 0) {
		return rand()%16;
	} 
	return wins;
}

static int
tr_nodes_depth(cboard_t *root, int depth)
{
	cboard_t *current = root;
	int count = 0;
	if (depth == 0) {
		count = 1;
		//printf("tr_nodes_depth: Depth = %d, Nodes = %d\n", depth, count);
		return 1;  /* root */
	}
    while (current != NULL) {
		if (current->n_bp[0]) {
			current = current->n_bp[0]; 
			if (current && current->b_depth == depth) {
				count++;
			}
		} else {
			/*
			 * Leaf node
			 */
			while (current->b_sibling == NULL && current != root) {
				current = current->b_parent;
			}
			current = current->b_sibling;
			if (current && current->b_depth == depth) {
				count++;
			}
		}
	}
#if 0
	if (count > 0) {
		printf("tr_nodes_depth: Depth = %d, Nodes = %d\n", depth, count);
	}
#endif
	return count;
}

/*
 * Average number of nodes visited at the given depth
 */
static int
tr_avg_nodes_depth(int depth)
{
	int i;
	int num_trees = 0;
	int total_nodes = 0;
	float avg_nodes = 0;
	int count = 0;

	for (i = 0; i < MAX_TREES && ntrees[i]; i++) {
		count = tr_nodes_depth(ntrees[i], depth);
		if (count > 0) {
			num_trees++;
		}
		total_nodes += count;
	}
	avg_nodes = (float)((float)total_nodes/(float)num_trees);
	if (sdbg) {
		printf("depth = %d, total_nodes = %d, num_trees = %d, avg_nodes = %f\n",
			depth, total_nodes, num_trees, avg_nodes);
	}
	printf("Depth = %d, Average nodes = %f\n", depth, avg_nodes);
	return;
}
	
/*
 * Code for hierarchical clustering
 */
#define MAX_ELEMS      MAX_TREES
#define MAX_ELEM_NAME  16384  /* Size of each elem name */
#define MAX_NODES      MAX_ELEMS*2
struct node
{
	char        elem[MAX_ELEM_NAME]; 
	int         level;     /* depth in tree from the top */
	int         dist;  /* distance between elements if clustered */
	struct node *child;
};
typedef struct node node_t;

static void hi_build_dist_matrix(void);
static void hi_find_min(int *row, int *col, int *min);
static void hi_merge_clusters(void);
static void hi_build_new_dist(int row, int col);
static void hi_print_dist_matrix(void);
static boolean_t hi_all_merged(void);
static void hi_print_tree(int level);
static void hi_build_init_nodes(void);
static boolean_t hi_row_null(int i);
static boolean_t hi_col_null(int j);
static void hi_build_elems(void);
static int hi_elem_dist(int i, int j);
static void hi_build_cluster(void);

static int  hi_dist[MAX_ELEMS][MAX_ELEMS];
static char hi_elems[MAX_ELEMS][MAX_ELEM_NAME];

static node_t hi_nodes[MAX_NODES];  /* Will store all the clusters */
static int    hi_index;  /* Next node will be allocated here */
static int    hi_level;  /* Next level for the tree */
static int    hi_num_elems;

/*
 * Used temporarily.
 */
static int hi_tdist[MAX_ELEMS][MAX_ELEMS];

static void
hi_merge_clusters(void)
{
	printf("\nMERGING HIERARCHICAL CLUSTERS\n\n");
	/*
	 * Fill in the elements and the nodes array initially
	 */
	hi_build_elems();

	/*
	 * Build the distance matrix
	 */
	hi_build_dist_matrix();

	/*
	 * Build the cluster now.
	 */
	hi_build_cluster();
	return;
}


static void
hi_build_elems(void)
{
	int i;

	for (i = 0; i < MAX_TREES && ntrees[i]; i++, hi_num_elems++) {
		cboard_t *root = ntrees[i];

		bzero(hi_nodes[i].elem, MAX_ELEM_NAME);

		sprintf(hi_elems[i], "%s_%d_%d", root->b_cmove, i, root->b_game);

		//sprintf(hi_elems[i], "%s_%d", root->b_cmove, i);

		/*
		 * Fill the node too
		 */
		strcpy(hi_nodes[i].elem, hi_elems[i]);
		hi_nodes[i].level = 0;
		hi_nodes[i].dist  =  0;
		hi_nodes[i].child = NULL;
		hi_index++;
	}
	hi_level++;

	if (sdbg) {
		printf("ELEMS: ");
		for (i = 0; i < hi_num_elems; i++) {
			printf("%s ", hi_elems[i]);
		}
		printf("\n");

		printf("NODES:  ");
		for(i = 0; i < hi_num_elems*2; i++) {
			if (hi_nodes[i].elem !="") {
				printf("%s ", hi_nodes[i].elem);
			}
		}
		printf("\n");
	}
	return;
}

static void
hi_build_dist_matrix(void)
{
	int i, j;
	bzero(hi_dist, sizeof(hi_dist));
	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			if (i != j && hi_dist[i][j] == 0) {
				int n = hi_elem_dist(i, j);
				hi_dist[i][j] = hi_dist[j][i] = n;
			}
		}
	}
	hi_print_dist_matrix();
	return;
}

static int
hi_elem_dist(int row, int col)
{
	//printf("Distance between %s:%s\n", hi_elems[row], hi_elems[col]);
	return tr_dist(row, col);
}
	

/*
 * Merge all the clusters one by one until we are done.
 */
static void
hi_build_cluster(void)
{
	int row, col, min;
	int i;

	/*
	 * Keep merging the clusters until we are done!
	 */
	for(;;) {
		printf("\n\nMERGING CLUSTER. Level = %d\n\n", hi_level);
		hi_find_min(&row, &col, &min);
		hi_build_new_dist(row, col);

		if (hi_all_merged() == TRUE) {
			printf("\nALL CLUSTERS MERGED SUCCESSFULLY!\n\n");
			break;
		}
	}

	for (i = 0; i < hi_num_elems; i++) {
		hi_print_tree(i);
	}
	return;
}

/*
 * Find min in the global gdist
 */
static void
hi_find_min(int *row, int *col, int *min)
{
	int i, j, tmin;
	tmin = 65536;
	*row = *col = *min = -1;
	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			if (i == j || hi_dist[i][j] == 0) {
				continue;
			}
			tmin = MIN(tmin, hi_dist[i][j]);
		}
	}
	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			if (hi_dist[i][j] == tmin) {
				*row = i;
				*col = j;
				*min = tmin;
				return;
			}
		}
	}
	return;
}

/*
 * This will modify the original global distance matrix
 */
static void
hi_build_new_dist(int row, int col)
{
	int i, j;
	int r1, r2;
	char elem[MAX_ELEM_NAME];
	node_t *np;
	int index1, index2;
	int distance = hi_dist[row][col];

	index1 = index2 = -1;

	//printf("\nBuilding new matrix\n");
	//printf("row = %d, col = %d, hi_num_elems = %d\n", row, col, hi_num_elems);
	if (sdbg) {
		printf("Closest %s:%s, %d\n", hi_elems[row], hi_elems[col],  hi_dist[row][col]);
	}
	printf("Closest %s:%s, %d\n", hi_elems[row], hi_elems[col],  hi_dist[row][col]);
	bzero(elem, MAX_ELEM_NAME);
	strcpy(elem, hi_elems[row]);
	strcat(elem, "/");
	strcat(elem, hi_elems[col]);

	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			hi_tdist[i][j] = hi_dist[i][j];
		}
	}
	i = MAX(row, col);
	for (j = 0; j < hi_num_elems; j++) {
		hi_tdist[i][j] = 0;
	}
	j = MAX(row, col);
	for (i = 0; i < hi_num_elems; i++) {
		hi_tdist[i][j] = 0;
	}
	r1 = MIN(row, col);
	r2 = MAX(row, col);
	i = r1;
	for (j = 0; j < hi_num_elems; j++) {
		hi_tdist[i][j] = MIN(hi_dist[r1][j], hi_dist[r2][j]);
		hi_tdist[j][i] = hi_tdist[i][j];
	}
	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			hi_dist[i][j] = hi_tdist[i][j];
		}
	}
	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			hi_tdist[i][j] = 0;
		}
	}

	np = &hi_nodes[hi_index];
	strcpy(np->elem, elem);
	np->level = hi_level;
	np->dist =  distance;
	np->child = NULL;

	/*
	 * Get the index based on the name
	 */
	for(i = 0; i < hi_num_elems*2; i++) {
		if (strcmp(hi_nodes[i].elem, hi_elems[row]) == 0) {
			index1 = i;
		}
		if (strcmp(hi_nodes[i].elem, hi_elems[col]) == 0) {
			index2 = i;
		}
		if (index1 >= 0 && index2 >= 0) {
			break;
		}
	}
	//printf("index1 = %d, index2 = %d\n", index1, index2);
	hi_nodes[index1].child = hi_nodes[index2].child = &hi_nodes[hi_index];

	if (sdbg) {
		printf("New elem = %s. Level = %d. Distance = %d\n",
		np->elem, np->level, np->dist);
	}
	hi_index++;
	hi_level++;

	/*
	 * Update the global element also for names
	 */
	for(i = 0; i < hi_num_elems; i++) {
		bzero(hi_elems[r1], MAX_ELEM_NAME);
		bzero(hi_elems[r2], MAX_ELEM_NAME);
		strcpy(hi_elems[r1], elem);
	}

	if (sdbg) {
		printf("ELEMENTS:  ");
		for(i = 0; i < hi_num_elems; i++) {
			printf("%s ", hi_elems[i]);
		}
		printf("\n");

		/*
		 * Nodes array
		 */
		printf("NODES:  ");
		for(i = 0; i < hi_num_elems*2; i++) {
			if (hi_nodes[i].elem !="") {
				printf("%s ", hi_nodes[i].elem);
			}
		}
		printf("\n");
	}
	hi_print_dist_matrix();
	return;
}


static boolean_t
hi_row_null(int i)
{
	int j;
	for (j = 0; j < hi_num_elems; j++) {
		if (hi_dist[i][j] != 0) {
			return FALSE;
		}
	}
	return TRUE;
}

static boolean_t
hi_col_null(int j)
{
	int i;
	for (i = 0; i < hi_num_elems; i++) {
		if (hi_dist[i][j] != 0) {
			return FALSE;
		}
	}
	return TRUE;
}

static boolean_t
hi_all_merged(void)
{
	int i;
	for (i = 0; i < hi_num_elems; i++) {
		if (hi_row_null(i) == FALSE) {
			return FALSE;
		}
	}
	return TRUE;
}

/*
 * Print the distance matrix
 */
static void
hi_print_dist_matrix(void)
{
	int i, j;

	if (sdbg == FALSE) {
		return;
	}
	printf("\nDISTANCE MATRIX\n");
	for (i = 0; i < hi_num_elems; i++) {
		for (j = 0; j < hi_num_elems; j++) {
			printf("%-10d ", hi_dist[i][j]);
		}
		printf("\n");
	}
	return;
}

static void
hi_print_tree(int level)
{
	int i;
	if (sdbg == FALSE) {
		return;
	}

	printf("\nPrinting tree at level %d\n", level);

	for (i = 0; i < hi_num_elems*2; i++) {
		if (hi_nodes[i].level > level) {
			continue;
		}
		if (hi_nodes[i].level == level) {
			printf("%s ", hi_nodes[i].elem);
			continue;
		}
		if (hi_nodes[i].level < level) {
			node_t *np = hi_nodes[i].child;
			if (np && np->level > level) {
				printf("%s ", hi_nodes[i].elem);
				continue;
			}
		}
	}
	printf("\n");
	return;
}

/*
 * CODE for book extensions
 */

#define PMAX_DEPTH 12

/*
 * Call from each depth of PVS search to see if there
 * is a matching book entry based upon the new board
 * position at this pvs move
 */
int
BookPVSQuery(HashType hkey, char *move, int score, int *new_score)
{
	int j;
	int found = false;
	int weight = 0;
	fflush(gdbg_fp);

	if (!(flags & BOOKPVS)) {
		return;
	}

	if (bookloaded && !book_allocated) {
		return false;
	}

	for (DIGEST_START(j,hkey); !DIGEST_EMPTY(j); DIGEST_NEXT(j, hkey)) {
		if (DIGEST_MATCH(j, hkey)) {
			if (bookpos[j].wins > 0) {
				printf("\nBOOKPVS match found HashKey = 0x%x ", hkey);
				printf(" move = %s wins = %d \n",
					move, bookpos[j].wins);
				fflush(gdbg_fp);
				found = true;
				break;
			}
		}
	}
	if (found == false) {
		return false;
	}

	/*
	 * Update the book score.
	 */
	if (flags & UPDATESCORE) {
		fprintf(gdbg_fp, "BookPVSScoreUpdate: Idepth = %d, ElapsedTime = %.2f, score = %d, maxtime = %.2f, SearchTime = %.2f\n",
		Idepth, ElapsedTime, score, maxtime, SearchTime);
	
		if (SearchTime > ElapsedTime) {
			weight = ((8192 * (int)(SearchTime - ElapsedTime))/(int) SearchTime) +
				(bookpos[j].wins - bookpos[j].losses + bookpos[j].draws/2) +
				(10 * (PMAX_DEPTH - Idepth)/PMAX_DEPTH);
			*new_score = score * weight;
			printf("Old score = %d, Weight = %d, New score = %d\n",
				score, weight, *new_score);
		} else {
			weight = (bookpos[j].wins - bookpos[j].losses + bookpos[j].draws/2) +
					(10 * (PMAX_DEPTH - Idepth)/PMAX_DEPTH);
			*new_score = score * weight;
			printf("Old score = %d, Weight = %d, New score = %d\n",
				score, weight, *new_score);
		}
	}
	fflush(gdbg_fp);
	fflush(stdout);
	return true;
}

void 
BookPVSScoreUpdate(int score, int *new_score)
{
   int i, len;
   int pvar[MAXPLYDEPTH];
   int found = false;

	if (!(flags & UPDATESCORE)) {
		return;
	}

   /* SMC */
   if (!(flags & POST))
     return;
   if (rootscore == -INFINITY-1)
      return;
   ElapsedTime = GetElapsed (StartTime);

   SANMove (RootPV, 1);
   MakeMove (board.side, &RootPV);

   TreePtr[3] = TreePtr[2];
   GenMoves (2);
   len = strlen (SANmv);
   i = 2;
   pvar[1] = RootPV;

	found = BookPVSQuery(HashKey, SANmv, score, new_score);
	if (found == true) {
#if 0
		SANMove (RootPV, 1);
		SET (flags, TIMEOUT);
		fprintf(gdbg_fp, "search.c TIMEOUT bookpvs set\n");
		fflush(gdbg_fp);
#endif
	}

   for (--i; i; i--)
      UnmakeMove (board.side, &pvar[i]);
   fflush (stdout);
   fflush (gdbg_fp);
   if (ofp != stdout) fflush (ofp);
}

/*
 * Call from each depth of PVS search to see if there
 * is a matching tree entry based upon the new board
 * position at this pvs move
 */
int
TreePVSQuery(HashType hkey, char *move, int score, int *new_score)
{
	int j;
	int found = false;
	int weight = 0;
	fflush(gdbg_fp);
	int i;
	cboard_t *bp;


	if (bookloaded && !book_allocated) {
		return FALSE;
	}

	for (i = 0; i < MAX_TREES; i++) {
		if (ntrees[i] == NULL) {
			return FALSE;
		}

		bp = tr_find_entry(ntrees[i], hkey);
		if (bp) {
			printf("\nTREEPVS match found HashKey = 0x%x ", hkey);
			printf(" move = %s wins = %d \n",
					move, bp->b_wins);
			found = TRUE;
			break;
		}
	}
	if (found == FALSE) {
		return FALSE;
	}

	if (flags & TUPDATESCORE) {
		fprintf(gdbg_fp, "TreePVSQuery: Idepth = %d, ElapsedTime = %.2f, score = %d, maxtime = %.2f, SearchTime = %.2f\n",
		Idepth, ElapsedTime, score, maxtime, SearchTime);
	
		if (SearchTime > ElapsedTime) {
			weight = ((8192 * (int)(SearchTime - ElapsedTime))/(int) SearchTime) +
				(bp->b_wins - bp->b_losses) +
				(10 * (PMAX_DEPTH - Idepth)/PMAX_DEPTH);
			*new_score = score * weight;
			printf("Old score = %d, Weight = %d, New score = %d\n",
				score, weight, *new_score);
		} else {
			weight = (bp->b_wins - bp->b_losses) +
					(10 * (PMAX_DEPTH - Idepth)/PMAX_DEPTH);
			*new_score = score * weight;
			printf("Old score = %d, Weight = %d, New score = %d\n",
				score, weight, *new_score);
		}
	}
	fflush(gdbg_fp);
	fflush(stdout);
	return TRUE;
}

void 
TreePVSScoreUpdate(int score, int *new_score)
{
   int i, len;
   int pvar[MAXPLYDEPTH];
   int found = false;

	if (!(flags & TUPDATESCORE)) {
		return;
	}

   /* SMC */
   if (!(flags & POST))
     return;
   if (rootscore == -INFINITY-1)
      return;
   ElapsedTime = GetElapsed (StartTime);

   SANMove (RootPV, 1);
   MakeMove (board.side, &RootPV);

   TreePtr[3] = TreePtr[2];
   GenMoves (2);
   len = strlen (SANmv);
   i = 2;
   pvar[1] = RootPV;

	found = TreePVSQuery(HashKey, SANmv, score, new_score);
	if (found == true) {
#if 0
		SANMove (RootPV, 1);
		SET (flags, TIMEOUT);
		fprintf(gdbg_fp, "search.c TIMEOUT bookpvs set\n");
		fflush(gdbg_fp);
#endif
	}

   for (--i; i; i--)
      UnmakeMove (board.side, &pvar[i]);
   fflush (stdout);
   fflush (gdbg_fp);
   if (ofp != stdout) fflush (ofp);
}
